"""
Base class for tensor-product style meshes
"""

import numpy as np
import scipy.sparse as sp
import properties

from .base_mesh import BaseMesh
from .. import utils

class BaseTensorMesh(BaseMesh):
    """
    Base class for tensor-product style meshes

    This class contains properites and methods that are common to cartesian
    and cylindrical meshes defined by tensor-produts of vectors describing
    cell spacings.

    Do not use this class directly, instead, inherit it if you plan to develop
    a tensor-style mesh (e.g. a spherical mesh) or use the
    :meth:`discretize.TensorMesh` class to create a cartesian tensor mesh.

    """
    _meshType = 'BASETENSOR'

    _unitDimensions = [1, 1, 1]

    # properties
    h = properties.List(
        "h is a list containing the cell widths of the tensor mesh in each "
        "dimension.",
        properties.Array(
            "widths of the tensor mesh in a single dimension",
            dtype=float,
            shape=("*",),
        ),
        max_length=3
    )

    def __init__(self, h=None, x0=None, **kwargs):

        h_in = h
        x0_in = x0

        # Sanity Checks
        assert type(h_in) in [list, tuple], 'h_in must be a list, not {}'.format(type(h_in))
        assert len(h_in) in [1, 2, 3], (
            'h_in must be of dimension 1, 2, or 3 not {}'.format(len(h_in))
        )

        # build h
        h = list(range(len(h_in)))
        for i, h_i in enumerate(h_in):
            if utils.isScalar(h_i) and type(h_i) is not np.ndarray:
                # This gives you something over the unit cube.
                h_i = self._unitDimensions[i] * np.ones(int(h_i))/int(h_i)
            elif type(h_i) is list:
                h_i = utils.meshTensor(h_i)
            assert isinstance(h_i, np.ndarray), (
                "h[{0:d}] is not a numpy array.".format(i)
            )
            assert len(h_i.shape) == 1, (
                "h[{0:d}] must be a 1D numpy array.".format(i)
            )
            h[i] = h_i[:]  # make a copy.

        # Origin of the mesh
        x0 = np.zeros(len(h))

        if x0_in is not None:
            assert len(h) == len(x0_in), "Dimension mismatch. x0 != len(h)"
            for i in range(len(h)):
                x_i, h_i = x0_in[i], h[i]
                if utils.isScalar(x_i):
                    x0[i] = x_i
                elif x_i == '0':
                    x0[i] = 0.0
                elif x_i == 'C':
                    x0[i] = -h_i.sum()*0.5
                elif x_i == 'N':
                    x0[i] = -h_i.sum()
                else:
                    raise Exception(
                        "x0[{0:d}] must be a scalar or '0' to be zero, "
                        "'C' to center, or 'N' to be negative. The input value"
                        " {1} {2} is invalid".format(i, x_i, type(x_i))
                    )

        if 'n' in kwargs.keys():
            n = kwargs.pop('n')
            assert (n == np.array([x.size for x in h])).all(), (
                "Dimension mismatch. The provided n doesn't "
            )
        else:
            n = np.array([x.size for x in h])

        super(BaseTensorMesh, self).__init__(
            n, x0=x0, **kwargs
        )

        # Ensure h contains 1D vectors
        self.h = [utils.mkvc(x.astype(float)) for x in h]

    @property
    def hx(self):
        """Width of cells in the x direction"""
        return self.h[0]

    @property
    def hy(self):
        "Width of cells in the y direction"
        return None if self.dim < 2 else self.h[1]

    @property
    def hz(self):
        "Width of cells in the z direction"
        return None if self.dim < 3 else self.h[2]

    @property
    def vectorNx(self):
        """Nodal grid vector (1D) in the x direction."""
        return np.r_[0., self.hx.cumsum()] + self.x0[0]

    @property
    def vectorNy(self):
        """Nodal grid vector (1D) in the y direction."""
        return (
            None if self.dim < 2 else np.r_[0., self.hy.cumsum()] + self.x0[1]
        )

    @property
    def vectorNz(self):
        """Nodal grid vector (1D) in the z direction."""
        return (
            None if self.dim < 3 else np.r_[0., self.hz.cumsum()] + self.x0[2]
        )

    @property
    def vectorCCx(self):
        """Cell-centered grid vector (1D) in the x direction."""
        return np.r_[0, self.hx[:-1].cumsum()] + self.hx*0.5 + self.x0[0]

    @property
    def vectorCCy(self):
        """Cell-centered grid vector (1D) in the y direction."""
        return (
            None if self.dim < 2 else
            np.r_[0, self.hy[:-1].cumsum()] + self.hy*0.5 + self.x0[1]
        )

    @property
    def vectorCCz(self):
        """Cell-centered grid vector (1D) in the z direction."""
        return (
            None if self.dim < 3 else
            np.r_[0, self.hz[:-1].cumsum()] + self.hz*0.5 + self.x0[2]
        )

    @property
    def gridCC(self):
        """Cell-centered grid."""
        return self._getTensorGrid('CC')

    @property
    def gridN(self):
        """Nodal grid."""
        return self._getTensorGrid('N')

    @property
    def h_gridded(self):
        """
        Returns an (nC, dim) numpy array with the widths of all cells in order
        """

        if self.dim == 1:
            return np.reshape(self.h, (self.nC, 1))
        elif self.dim == 2:
            hx = np.kron(np.ones(self.nCy), self.h[0])
            hy = np.kron(self.h[1], np.ones(self.nCx))
            return np.c_[hx, hy]
        elif self.dim == 3:
            hx = np.kron(np.ones(self.nCy*self.nCz), self.h[0])
            hy = np.kron(np.ones(self.nCz), np.kron(self.h[1], np.ones(self.nCx)))
            hz = np.kron(self.h[2], np.ones(self.nCx*self.nCy))
            return np.c_[hx, hy, hz]

    @property
    def gridFx(self):
        """Face staggered grid in the x direction."""
        if self.nFx == 0:
            return
        return self._getTensorGrid('Fx')

    @property
    def gridFy(self):
        """Face staggered grid in the y direction."""
        if self.nFy == 0 or self.dim < 2:
            return
        return self._getTensorGrid('Fy')

    @property
    def gridFz(self):
        """Face staggered grid in the z direction."""
        if self.nFz == 0 or self.dim < 3:
            return
        return self._getTensorGrid('Fz')

    @property
    def gridEx(self):
        """Edge staggered grid in the x direction."""
        if self.nEx == 0:
            return
        return self._getTensorGrid('Ex')

    @property
    def gridEy(self):
        """Edge staggered grid in the y direction."""
        if self.nEy == 0 or self.dim < 2:
            return
        return self._getTensorGrid('Ey')

    @property
    def gridEz(self):
        """Edge staggered grid in the z direction."""
        if self.nEz == 0 or self.dim < 3:
            return
        return self._getTensorGrid('Ez')

    def _getTensorGrid(self, key):
        if getattr(self, '_grid' + key, None) is None:
            setattr(self, '_grid' + key, utils.ndgrid(self.getTensor(key)))
        return getattr(self, '_grid' + key)

    def getTensor(self, key):
        """ Returns a tensor list.

        Parameters
        ----------

        key : str
            Which tensor (see below)

            key can be::

                'CC'    -> scalar field defined on cell centers
                'N'     -> scalar field defined on nodes
                'Fx'    -> x-component of field defined on faces
                'Fy'    -> y-component of field defined on faces
                'Fz'    -> z-component of field defined on faces
                'Ex'    -> x-component of field defined on edges
                'Ey'    -> y-component of field defined on edges
                'Ez'    -> z-component of field defined on edges

        Returns
        -------

        list
            list of the tensors that make up the mesh.

        """

        if key == 'Fx':
            ten = [self.vectorNx, self.vectorCCy, self.vectorCCz]
        elif key == 'Fy':
            ten = [self.vectorCCx, self.vectorNy, self.vectorCCz]
        elif key == 'Fz':
            ten = [self.vectorCCx, self.vectorCCy, self.vectorNz]
        elif key == 'Ex':
            ten = [self.vectorCCx, self.vectorNy, self.vectorNz]
        elif key == 'Ey':
            ten = [self.vectorNx, self.vectorCCy, self.vectorNz]
        elif key == 'Ez':
            ten = [self.vectorNx, self.vectorNy, self.vectorCCz]
        elif key == 'CC':
            ten = [self.vectorCCx, self.vectorCCy, self.vectorCCz]
        elif key == 'N':
            ten = [self.vectorNx, self.vectorNy, self.vectorNz]

        return [t for t in ten if t is not None]

    # --------------- Methods ---------------------

    def isInside(self, pts, locType='N'):
        """
        Determines if a set of points are inside a mesh.

        :param numpy.ndarray pts: Location of points to test
        :rtype: numpy.ndarray
        :return: inside, numpy array of booleans
        """
        pts = utils.asArray_N_x_Dim(pts, self.dim)

        tensors = self.getTensor(locType)

        if locType == 'N' and self._meshType == 'CYL':
            # NOTE: for a CYL mesh we add a node to check if we are inside in
            # the radial direction!
            tensors[0] = np.r_[0., tensors[0]]
            tensors[1] = np.r_[tensors[1], 2.0*np.pi]

        inside = np.ones(pts.shape[0], dtype=bool)
        for i, tensor in enumerate(tensors):
            TOL = np.diff(tensor).min() * 1.0e-10
            inside = (
                inside &
                (pts[:, i] >= tensor.min()-TOL) &
                (pts[:, i] <= tensor.max()+TOL)
            )
        return inside

    def _getInterpolationMat(self, loc, locType='CC', zerosOutside=False):
        """ Produces interpolation matrix

        Parameters
        ----------
        loc : numpy.ndarray
            Location of points to interpolate to

        locType: stc
            What to interpolate

            locType can be::

                'Ex'    -> x-component of field defined on edges
                'Ey'    -> y-component of field defined on edges
                'Ez'    -> z-component of field defined on edges
                'Fx'    -> x-component of field defined on faces
                'Fy'    -> y-component of field defined on faces
                'Fz'    -> z-component of field defined on faces
                'N'     -> scalar field defined on nodes
                'CC'    -> scalar field defined on cell centers
                'CCVx'  -> x-component of vector field defined on cell centers
                'CCVy'  -> y-component of vector field defined on cell centers
                'CCVz'  -> z-component of vector field defined on cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            M, the interpolation matrix

        """

        loc = utils.asArray_N_x_Dim(loc, self.dim)

        if not zerosOutside:
            assert np.all(self.isInside(loc)), "Points outside of mesh"
        else:
            indZeros = np.logical_not(self.isInside(loc))
            loc[indZeros, :] = np.array([
                v.mean() for v in self.getTensor('CC')
            ])

        if locType in ['Fx', 'Fy', 'Fz', 'Ex', 'Ey', 'Ez']:
            ind = {'x': 0, 'y': 1, 'z': 2}[locType[1]]
            assert self.dim >= ind, 'mesh is not high enough dimension.'
            nF_nE = self.vnF if 'F' in locType else self.vnE
            components = [utils.spzeros(loc.shape[0], n) for n in nF_nE]
            components[ind] = utils.interpmat(loc, *self.getTensor(locType))
            # remove any zero blocks (hstack complains)
            components = [comp for comp in components if comp.shape[1] > 0]
            Q = sp.hstack(components)

        elif locType in ['CC', 'N']:
            Q = utils.interpmat(loc, *self.getTensor(locType))

        elif locType in ['CCVx', 'CCVy', 'CCVz']:
            Q = utils.interpmat(loc, *self.getTensor('CC'))
            Z = utils.spzeros(loc.shape[0], self.nC)
            if locType == 'CCVx':
                Q = sp.hstack([Q, Z, Z])
            elif locType == 'CCVy':
                Q = sp.hstack([Z, Q, Z])
            elif locType == 'CCVz':
                Q = sp.hstack([Z, Z, Q])

        else:
            raise NotImplementedError(
                'getInterpolationMat: locType==' + locType +
                ' and mesh.dim==' + str(self.dim)
            )

        if zerosOutside:
            Q[indZeros, :] = 0

        return Q.tocsr()

    def getInterpolationMat(self, loc, locType='CC', zerosOutside=False):
        """ Produces interpolation matrix

        Parameters
        ----------
        loc : numpy.ndarray
            Location of points to interpolate to

        locType : str
            What to interpolate (see below)

            locType can be::

                'Ex'    -> x-component of field defined on edges
                'Ey'    -> y-component of field defined on edges
                'Ez'    -> z-component of field defined on edges
                'Fx'    -> x-component of field defined on faces
                'Fy'    -> y-component of field defined on faces
                'Fz'    -> z-component of field defined on faces
                'N'     -> scalar field defined on nodes
                'CC'    -> scalar field defined on cell centers
                'CCVx'  -> x-component of vector field defined on cell centers
                'CCVy'  -> y-component of vector field defined on cell centers
                'CCVz'  -> z-component of vector field defined on cell centers

        Returns
        -------

        scipy.sparse.csr_matrix
            M, the interpolation matrix

        """
        return self._getInterpolationMat(loc, locType, zerosOutside)

    def _fastInnerProduct(
        self, projType, prop=None, invProp=False, invMat=False
    ):
        """ Fast version of getFaceInnerProduct.
            This does not handle the case of a full tensor prop.

        Parameters
        ----------

        prop : numpy.array
            material property (tensor properties are possible) at each cell center (nC, (1, 3, or 6))

        projType : str
            'E' or 'F'

        returnP : bool
            returns the projection matrices

        invProp : bool
            inverts the material property

        invMat : bool
            inverts the matrix

        Returns
        -------
        scipy.sparse.csr_matrix
            M, the inner product matrix (nF, nF)

        """
        assert projType in ['F', 'E'], (
            "projType must be 'F' for faces or 'E' for edges"
        )

        if prop is None:
            prop = np.ones(self.nC)

        if invProp:
            prop = 1./prop

        if utils.isScalar(prop):
            prop = prop*np.ones(self.nC)

        # number of elements we are averaging (equals dim for regular
        # meshes, but for cyl, where we use symmetry, it is 1 for edge
        # variables and 2 for face variables)
        if self._meshType == 'CYL':
            n_elements = np.sum(getattr(self, 'vn'+projType).nonzero())
        else:
            n_elements = self.dim

        # Isotropic? or anisotropic?
        if prop.size == self.nC:
            Av = getattr(self, 'ave'+projType+'2CC')
            Vprop = self.vol * utils.mkvc(prop)
            M = n_elements * utils.sdiag(Av.T * Vprop)

        elif prop.size == self.nC*self.dim:
            Av = getattr(self, 'ave'+projType+'2CCV')

            # if cyl, then only certain components are relevant due to symmetry
            # for faces, x, z matters, for edges, y (which is theta) matters
            if self._meshType == 'CYL':
                if projType == 'E':
                    prop = prop[:, 1]  # this is the action of a projection mat
                elif projType == 'F':
                    prop = prop[:, [0, 2]]

            V = sp.kron(sp.identity(n_elements), utils.sdiag(self.vol))
            M = utils.sdiag(Av.T * V * utils.mkvc(prop))
        else:
            return None

        if invMat:
            return utils.sdInv(M)
        else:
            return M

    def _fastInnerProductDeriv(self, projType, prop, invProp=False,
                               invMat=False):
        """

        Parameters
        ----------

        projType : str
            'E' or 'F'

        tensorType : TensorType
            type of the tensor

        invProp : bool
            inverts the material property

        invMat : bool
            inverts the matrix


        Returns
        -------
        function
            dMdmu, the derivative of the inner product matrix

        """
        assert projType in ['F', 'E'], ("projType must be 'F' for faces or 'E'"
                                        " for edges")

        tensorType = utils.TensorType(self, prop)

        dMdprop = None

        if invMat or invProp:
            MI = self._fastInnerProduct(projType, prop, invProp=invProp,
                                        invMat=invMat)

        # number of elements we are averaging (equals dim for regular
        # meshes, but for cyl, where we use symmetry, it is 1 for edge
        # variables and 2 for face variables)
        if self._meshType == 'CYL':
            n_elements = np.sum(getattr(self, 'vn'+projType).nonzero())
        else:
            n_elements = self.dim

        if tensorType == 0:  # isotropic, constant
            Av = getattr(self, 'ave'+projType+'2CC')
            V = utils.sdiag(self.vol)
            ones = sp.csr_matrix(
                (np.ones(self.nC), (range(self.nC), np.zeros(self.nC))),
                shape=(self.nC, 1)
            )
            if not invMat and not invProp:
                dMdprop = n_elements * Av.T * V * ones
            elif invMat and invProp:
                dMdprop =  n_elements * (
                    utils.sdiag(MI.diagonal()**2) * Av.T * V * ones *
                    utils.sdiag(1./prop**2)
                )
            elif invProp:
                dMdprop = n_elements * Av.T * V * utils.sdiag(- 1./prop**2)
            elif invMat:
                dMdprop = n_elements * (
                    utils.sdiag(- MI.diagonal()**2) * Av.T * V
                )

        elif tensorType == 1:  # isotropic, variable in space
            Av = getattr(self, 'ave'+projType+'2CC')
            V = utils.sdiag(self.vol)
            if not invMat and not invProp:
                dMdprop = n_elements * Av.T * V
            elif invMat and invProp:
                dMdprop =  n_elements * (
                    utils.sdiag(MI.diagonal()**2) * Av.T * V *
                    utils.sdiag(1./prop**2)
                )
            elif invProp:
                dMdprop = n_elements * Av.T * V * utils.sdiag(-1./prop**2)
            elif invMat:
                dMdprop = n_elements * (
                    utils.sdiag(- MI.diagonal()**2) * Av.T * V
                )

        elif tensorType == 2: # anisotropic
            Av = getattr(self, 'ave'+projType+'2CCV')
            V = sp.kron(sp.identity(self.dim), utils.sdiag(self.vol))

            if self._meshType == 'CYL':
                Zero = sp.csr_matrix((self.nC, self.nC))
                Eye = sp.eye(self.nC)
                if projType == 'E':
                    P = sp.hstack([Zero, Eye, Zero])
                    # print(P.todense())
                elif projType == 'F':
                    P = sp.vstack([sp.hstack([Eye, Zero, Zero]),
                                   sp.hstack([Zero, Zero, Eye])])
                    # print(P.todense())
            else:
                P = sp.eye(self.nC*self.dim)

            if not invMat and not invProp:
                dMdprop = Av.T * P * V
            elif invMat and invProp:
                dMdprop = (utils.sdiag(MI.diagonal()**2) * Av.T * P * V *
                           utils.sdiag(1./prop**2))
            elif invProp:
                dMdprop = Av.T * P * V * utils.sdiag(-1./prop**2)
            elif invMat:
                dMdprop = utils.sdiag(- MI.diagonal()**2) * Av.T * P * V

        if dMdprop is not None:
            def innerProductDeriv(v=None):
                if v is None:
                    warnings.warn(
                        "Depreciation Warning: TensorMesh.innerProductDeriv."
                        " You should be supplying a vector. "
                        "Use: sdiag(u)*dMdprop", FutureWarning
                    )
                    return dMdprop
                return utils.sdiag(v) * dMdprop
            return innerProductDeriv
        else:
            return None



import numpy as np
from scipy import sparse as sp
from .matutils import mkvc, ndgrid, sub2ind, sdiag


def volTetra(xyz, A, B, C, D):
    """
    Returns the volume for tetrahedras volume specified by the indexes A to D.

    :param numpy.ndarray xyz: X,Y,Z vertex vector
    :param numpy.ndarray A,B,C,D: vert index of the tetrahedra
    :rtype: numpy.ndarray
    :return: V, volume of the tetrahedra

    Algorithm https://en.wikipedia.org/wiki/Tetrahedron#Volume

    .. math::

       V = {1 \over 3} A  h

       V = {1 \over 6} | ( a - d ) \cdot ( ( b - d )  ( c - d ) ) |

    """

    AD = xyz[A, :] - xyz[D, :]
    BD = xyz[B, :] - xyz[D, :]
    CD = xyz[C, :] - xyz[D, :]

    V = (BD[:, 0]*CD[:, 1] - BD[:, 1]*CD[:, 0])*AD[:, 2] - (BD[:, 0]*CD[:, 2] - BD[:, 2]*CD[:, 0])*AD[:, 1] + (BD[:, 1]*CD[:, 2] - BD[:, 2]*CD[:, 1])*AD[:, 0]
    return V/6


def indexCube(nodes, gridSize, n=None):
    """
    Returns the index of nodes on the mesh.


    Input:
       nodes     - string of which nodes to return. e.g. 'ABCD'
       gridSize  - size of the nodal grid
       n         - number of nodes each i,j,k direction: [ni,nj,nk]


    Output:
       index  - index in the order asked e.g. 'ABCD' --> (A,B,C,D)

    TWO DIMENSIONS::

      node(i,j)          node(i,j+1)
           A -------------- B
           |                |
           |    cell(i,j)   |
           |        I       |
           |                |
          D -------------- C
      node(i+1,j)        node(i+1,j+1)


    THREE DIMENSIONS::

            node(i,j,k+1)       node(i,j+1,k+1)
                E --------------- F
               /|               / |
              / |              /  |
             /  |             /   |
      node(i,j,k)         node(i,j+1,k)
           A -------------- B     |
           |    H ----------|---- G
           |   /cell(i,j)   |   /
           |  /     I       |  /
           | /              | /
           D -------------- C
      node(i+1,j,k)      node(i+1,j+1,k)

    """

    assert type(nodes) == str, "Nodes must be a str variable: e.g. 'ABCD'"
    assert isinstance(gridSize, np.ndarray), "Number of nodes must be an ndarray"
    nodes = nodes.upper()
    # Make sure that we choose from the possible nodes.
    possibleNodes = 'ABCD' if gridSize.size == 2 else 'ABCDEFGH'
    for node in nodes:
        assert node in possibleNodes, "Nodes must be chosen from: '{0!s}'".format(possibleNodes)
    dim = gridSize.size
    if n is None:
        n = gridSize - 1

    if dim == 2:
        ij = ndgrid(np.arange(n[0]), np.arange(n[1]))
        i, j = ij[:, 0], ij[:, 1]
    elif dim == 3:
        ijk = ndgrid(np.arange(n[0]), np.arange(n[1]), np.arange(n[2]))
        i, j, k = ijk[:, 0], ijk[:, 1], ijk[:, 2]
    else:
        raise Exception('Only 2 and 3 dimensions supported.')

    nodeMap = {'A': [0, 0, 0], 'B': [0, 1, 0], 'C': [1, 1, 0], 'D': [1, 0, 0],
               'E': [0, 0, 1], 'F': [0, 1, 1], 'G': [1, 1, 1], 'H': [1, 0, 1]}
    out = ()
    for node in nodes:
        shift = nodeMap[node]
        if dim == 2:
            out += (sub2ind(gridSize, np.c_[i+shift[0], j+shift[1]]).flatten(), )
        elif dim == 3:
            out += (sub2ind(gridSize, np.c_[i+shift[0], j+shift[1], k+shift[2]]).flatten(), )

    return out


def faceInfo(xyz, A, B, C, D, average=True, normalizeNormals=True):
    """
    function [N] = faceInfo(y,A,B,C,D)

       Returns the averaged normal, area, and edge lengths for a given set of faces.

       If average option is FALSE then N is a cell array {nA,nB,nC,nD}


    Input:
       xyz          - X,Y,Z vertex vector
       A,B,C,D      - vert index of the face (counter clockwize)

    Options:
       average      - [true]/false, toggles returning all normals or the average

    Output:
       N            - average face normal or {nA,nB,nC,nD} if average = false
       area         - average face area
       edgeLengths  - exact edge Lengths, 4 column vector [AB, BC, CD, DA]

    see also testFaceNormal testFaceArea

    @author Rowan Cockett

    Last modified on: 2013/07/26

    """
    assert type(average) is bool, 'average must be a boolean'
    assert type(normalizeNormals) is bool, 'normalizeNormals must be a boolean'
    # compute normal that is pointing away from you.
    #
    #    A -------A-B------- B
    #    |                   |
    #    |                   |
    #   D-A       (X)       B-C
    #    |                   |
    #    |                   |
    #    D -------C-D------- C

    AB = xyz[B, :] - xyz[A, :]
    BC = xyz[C, :] - xyz[B, :]
    CD = xyz[D, :] - xyz[C, :]
    DA = xyz[A, :] - xyz[D, :]

    def cross(X, Y):
        return np.c_[X[:, 1]*Y[:, 2] - X[:, 2]*Y[:, 1],
                     X[:, 2]*Y[:, 0] - X[:, 0]*Y[:, 2],
                     X[:, 0]*Y[:, 1] - X[:, 1]*Y[:, 0]]

    nA = cross(AB, DA)
    nB = cross(BC, AB)
    nC = cross(CD, BC)
    nD = cross(DA, CD)

    length = lambda x: np.sqrt(x[:, 0]**2 + x[:, 1]**2 + x[:, 2]**2)
    normalize = lambda x: x/np.kron(np.ones((1, x.shape[1])), mkvc(length(x), 2))
    if average:
        # average the normals at each vertex.
        N = (nA + nB + nC + nD)/4  # this is intrinsically weighted by area
        # normalize
        N = normalize(N)
    else:
        if normalizeNormals:
            N = [normalize(nA), normalize(nB), normalize(nC), normalize(nD)]
        else:
            N = [nA, nB, nC, nD]

    # Area calculation
    #
    # Approximate by 4 different triangles, and divide by 2.
    # Each triangle is one half of the length of the cross product
    #
    # So also could be viewed as the average parallelogram.
    #
    # TODO: This does not compute correctly for concave quadrilaterals
    area = (length(nA)+length(nB)+length(nC)+length(nD))/4

    return N, area



from __future__ import print_function
import numpy as np
import scipy.sparse as sp
from .matutils import mkvc, sub2ind

try:
    from . import interputils_cython as pyx
    _interp_point_1D = pyx._interp_point_1D
    _interpmat1D = pyx._interpmat1D
    _interpmat2D = pyx._interpmat2D
    _interpmat3D = pyx._interpmat3D
    _vol_interp = pyx._tensor_volume_averaging
    _interpCython = True
except ImportError as err:
    print(err)
    import os
    # Check if being called from non-standard location (i.e. a git repository)
    # is tree_ext.cpp here? will not be in the folder if installed to site-packages...
    file_test = os.path.dirname(os.path.abspath(__file__))+"/interputils_cython.pyx"
    if os.path.isfile(file_test):
        # Then we are being run from a repository
        print(
            """
            Unable to import interputils_cython.

            It would appear that discretize is being imported from its repository.
            If this is intentional, you need to run:

            python setup.py build_ext --inplace

            to build the cython code.
            """
            )
    _interpCython = False


def interpmat(locs, x, y=None, z=None):
    """Local interpolation computed for each receiver point in turn

    :param numpy.ndarray loc: Location of points to interpolate to
    :param numpy.ndarray x: Tensor of 1st dimension of grid.
    :param numpy.ndarray y: Tensor of 2nd dimension of grid. None by default.
    :param numpy.ndarray z: Tensor of 3rd dimension of grid. None by default.
    :rtype: scipy.sparse.csr_matrix
    :return: Interpolation matrix

    .. plot::

        import discretize
        import numpy as np
        import matplotlib.pyplot as plt
        locs = np.random.rand(50)*0.8+0.1
        x = np.linspace(0, 1, 7)
        dense = np.linspace(0, 1, 200)
        fun = lambda x: np.cos(2*np.pi*x)
        Q = discretize.utils.interpmat(locs, x)
        plt.plot(x, fun(x), 'bs-')
        plt.plot(dense, fun(dense), 'y:')
        plt.plot(locs, Q*fun(x), 'mo')
        plt.plot(locs, fun(locs), 'rx')
        plt.show()

    """

    npts = locs.shape[0]
    locs = locs.astype(float)
    x = x.astype(float)
    if y is None and z is None:
        shape = [x.size]
        inds, vals = _interpmat1D(mkvc(locs), x)
    elif z is None:
        y = y.astype(float)
        shape = [x.size, y.size]
        inds, vals = _interpmat2D(locs, x, y)
    else:
        y = y.astype(float)
        z = z.astype(float)
        shape = [x.size, y.size, z.size]
        inds, vals = _interpmat3D(locs, x, y, z)

    I = np.repeat(range(npts), 2**len(shape))
    J = sub2ind(shape, inds)
    Q = sp.csr_matrix((vals, (I, J)),
                      shape=(npts, np.prod(shape)))
    return Q

def volume_average(mesh_in, mesh_out, values=None, output=None):
    """Volume averaging interpolation between meshes.

    This volume averaging function looks for overlapping cells in each mesh,
    and weights the output values by the partial volume ratio of the overlapping
    input cells. The volume average operation should result in an output such that
    ``np.sum(mesh_in.vol*values)`` = ``np.sum(mesh_out.vol*output)``, when the input
    and output meshes have the exact same extent. When the output mesh extent goes
    beyond the input mesh, it is assumed to have constant values in that direction.
    When the output mesh extent is smaller than the input mesh, only the overlapping
    extent of the input mesh contributes to the output.

    This function operates in three different modes. If only ``mesh_in`` and
    ``mesh_out`` are given, the returned value is a ``scipy.sparse.csr_matrix``
    that represents this operation (so it could potentially be applied repeatedly).
    If ``values`` is given, the volume averaging is performed right away (without
    internally forming the matrix) and the returned value is the result of this.
    If ``output`` is given as well, it will be filled with the values of the
    operation and then returned (assuming it has the correct ``dtype``).

    Parameters
    ----------
    mesh_in: TensorMesh or TreeMesh
        Input mesh (the mesh you are interpolating from)
    mesh_out: TensorMesh or TreeMesh
        Output mesh (the mesh you are interpolating to)
    values: numpy.ndarray, optional
        Array with values defined at the cells of ``mesh_in``
    output: numpy.ndarray, optional
        Output array to be overwritten of length ``mesh_out.nC`` and type ``np.float64``

    Returns
    -------
    scipy.sparse.csr_matrix or numpy.ndarray
        If ``values`` is ``None``, the returned value is a matrix representing this operation,
        otherwise it is a numpy.ndarray of the result of the operation.

    Examples
    --------
    Create two meshes with the same extent, but different divisions (the meshes
    do not have to be the same extent).

    >>> import numpy as np
    >>> from discretize import TensorMesh
    >>> h1 = np.ones(32)
    >>> h2 = np.ones(16)*2
    >>> mesh_in = TensorMesh([h1, h1])
    >>> mesh_out = TensorMesh([h2, h2])

    Create a random model defined on the input mesh, and use volume averaging to
    interpolate it to the output mesh.

    >>> from discretize.utils import volume_average
    >>> model1 = np.random.rand(mesh_in.nC)
    >>> model2 = volume_average(mesh_in, mesh_out, model1)

    Because these two meshes' cells are perfectly aligned, but the output mesh
    has 1 cell for each 4 of the input cells, this operation should effectively
    look like averaging each of those cells values

    >>> import matplotlib.pyplot as plt
    >>> plt.figure()
    >>> ax1 = plt.subplot(121)
    >>> mesh_in.plotImage(model1, ax=ax1)
    >>> ax2 = plt.subplot(122)
    >>> mesh_out.plotImage(model2, ax=ax2)
    >>> plt.show()

    """
    try:
        in_type = mesh_in._meshType
        out_type = mesh_out._meshType
    except AttributeError:
        raise TypeError("Both input and output mesh must be valid discetize meshes")

    valid_meshs = ['TENSOR', 'TREE']
    if in_type not in valid_meshs or out_type not in valid_meshs:
        raise NotImplementedError(f"Volume averaging is only implemented for TensorMesh and TreeMesh, "
                                  f"not {type(mesh_in).__name__} and/or {type(mesh_out).__name__}")

    if mesh_in.dim != mesh_out.dim:
        raise ValueError("Both meshes must have the same dimension")

    if values is not None and len(values) != mesh_in.nC:
        raise ValueError("Input array does not have the same length as the number of cells in input mesh")
    if output is not None and len(output) != mesh_out.nC:
        raise ValueError("Output array does not have the same length as the number of cells in output mesh")

    if values is not None:
        values = np.asarray(values, dtype=np.float64)
    if output is not None:
        output = np.asarray(output, dtype=np.float64)

    if in_type == 'TENSOR':
        if out_type == 'TENSOR':
            return _vol_interp(mesh_in, mesh_out, values, output)
        elif out_type == 'TREE':
            return mesh_out._vol_avg_from_tens(mesh_in, values, output)
    elif in_type == 'TREE':
        if out_type == 'TENSOR':
            return mesh_in._vol_avg_to_tens(mesh_out, values, output)
        elif out_type == 'TREE':
            return mesh_out._vol_avg_from_tree(mesh_in, values, output)
    else:
        raise TypeError("Unsupported mesh types")



from __future__ import print_function

import numpy as np
import properties
import scipy.sparse as sp
from scipy.constants import pi

from .utils import (
    kron3, ndgrid, av, speye, ddx, sdiag, interpmat, spzeros, cyl2cart
)
from .TensorMesh import BaseTensorMesh, BaseRectangularMesh
from .InnerProducts import InnerProducts
from .View import CylView
from .DiffOperators import DiffOperators


class CylMesh(
    BaseTensorMesh, BaseRectangularMesh, InnerProducts, CylView, DiffOperators
):
    """
    CylMesh is a mesh class for cylindrical problems. It supports both
    cylindrically symmetric and 3D cylindrical meshes that include an azimuthal
    discretization.

    For a cylindrically symmetric mesh use :code:`h = [hx, 1, hz]`. For example:

    .. plot::
        :include-source:

        import discretize
        from discretize import utils

        cs, nc, npad = 20., 30, 8
        hx = utils.meshTensor([(cs, npad+10, -0.7), (cs, nc), (cs, npad, 1.3)])
        hz = utils.meshTensor([(cs, npad ,-1.3), (cs, nc), (cs, npad, 1.3)])
        mesh = discretize.CylMesh([hx, 1, hz], x0=[0, 0, -hz.sum()/2])
        mesh.plotGrid()

    To create a 3D cylindrical mesh, we also include an azimuthal discretization

    .. plot::
        :include-source:

        import discretize
        from discretize import utils

        cs, nc, npad = 20., 30, 8
        nc_theta = 8
        hx = utils.meshTensor([(cs, npad+10, -0.7), (cs, nc), (cs, npad, 1.3)])
        hy = 2 * np.pi/nc_theta * np.ones(nc_theta)
        hz = utils.meshTensor([(cs,npad, -1.3), (cs,nc), (cs, npad, 1.3)])
        mesh = discretize.CylMesh([hx, hy, hz], x0=[0, 0, -hz.sum()/2])
        mesh.plotGrid()

    """

    _meshType = 'CYL'
    _unitDimensions = [1, 2*np.pi, 1]

    cartesianOrigin = properties.Array(
        "Cartesian origin of the mesh",
        dtype=float,
        shape=('*',)
    )

    def __init__(self, h=None, x0=None, **kwargs):
        super(CylMesh, self).__init__(h=h, x0=x0, **kwargs)
        self.reference_system = 'cylindrical'

        if not np.abs(self.hy.sum() - 2*np.pi) < 1e-10:
            raise AssertionError("The 2nd dimension must sum to 2*pi")

        if self.dim == 2:
            print('Warning, a disk mesh has not been tested thoroughly.')

        if 'cartesianOrigin' in kwargs.keys():
            self.cartesianOrigin = kwargs.pop('cartesianOrigin')
        else:
            self.cartesianOrigin = np.zeros(self.dim)

    @properties.validator('cartesianOrigin')
    def check_cartesian_origin_shape(self, change):
        change['value'] = np.array(change['value'], dtype=float).ravel()
        if len(change['value']) != self.dim:
            raise Exception(
                "Dimension mismatch. The mesh dimension is {}, and the "
                "cartesianOrigin provided has length {}".format(
                    self.dim, len(change['value'])
                )
            )

    @property
    def isSymmetric(self):
        """
        Is the mesh cylindrically symmetric?

        Returns
        -------
        bool
            True if the mesh is cylindrically symmetric, False otherwise
        """
        return self.nCy == 1

    @property
    def _ntNx(self):
        """
        Returns
        -------
        int
            Number of total x nodes (prior to deflating)
        """
        if self.isSymmetric:
            return self.nCx
        return self.nCx + 1

    @property
    def nNx(self):
        """
        Returns
        -------
        int
            Number of nodes in the x-direction
        """
        # if self.isSymmetric:
        #     return self.nCx
        return self._ntNx

    @property
    def _ntNy(self):
        """
        Returns
        -------
        int
            Number of total y nodes (prior to deflating)
        """
        if self.isSymmetric:
            return 1
        return self.nCy + 1

    @property
    def nNy(self):
        """
        Returns
        -------
        int
            Number of nodes in the y-direction
        """
        if self.isSymmetric:
            return 0
        return self.nCy

    @property
    def _ntNz(self):
        """
        Returns
        -------
        int
            Number of total z nodes (prior to deflating)
        """
        return self.nNz

    @property
    def _ntN(self):
        """
        Returns
        -------
        int
            Number of total nodes (prior to deflating)
        """
        if self.isSymmetric:
            return 0
        return int(self._ntNx * self._ntNy * self._ntNz)

    @property
    def nN(self):
        """
        Returns
        -------
        int
            Total number of nodes
        """
        if self.isSymmetric:
            return 0
        return (self.nNx - 1) * self.nNy * self.nNz + self.nNz

    @property
    def _vntFx(self):
        """
        vector number of total Fx (prior to deflating)
        """
        # if self.isSymmetric:
        #     return np.r_[self._ntNx, 1, self.nCz]
        return np.r_[self._ntNx, self.nCy, self.nCz]

    @property
    def _ntFx(self):
        """
        number of total Fx (prior to defplating)
        """
        return int(self._vntFx.prod())

    @property
    def _nhFx(self):
        """
        Number of hanging Fx
        """
        return int(self.nCy * self.nCz)

    @property
    def vnFx(self):
        """
        Returns
        -------
        numpy.ndarray
            Number of x-faces in each direction, (dim, )
        """
        return self.vnC

    @property
    def _vntFy(self):
        """
        vector number of total Fy (prior to deflating)
        """
        # if self.isSymmetric:
        #     return np.r_[0, 0, 0]
        return np.r_[self.nCx, self._ntNy, self.nCz]

    @property
    def _ntFy(self):
        """
        number of total Fy (prior to deflating)
        """
        return int(self._vntFy.prod())

    @property
    def _nhFy(self):
        """
        number of hanging y-faces
        """
        return int(self.nCx * self.nCz)

    @property
    def _vntFz(self):
        """
        vector number of total Fz (prior to deflating)
        """
        # if self.isSymmetric:
        #     return np.r_[self.nCx, 1, self._ntNz]
        return np.r_[self.nCx, self.nCy, self._ntNz]

    @property
    def _ntFz(self):
        """
        number of total Fz (prior to deflating)
        """
        return int(self._vntFz.prod())

    @property
    def _nhFz(self):
        """
        number of hanging Fz
        """
        return int(self.nCx * self.nCz)

    @property
    def _vntEx(self):
        """
        vector number of total Ex (prior to deflating)
        """
        return np.r_[self.nCx, self._ntNy, self._ntNz]

    @property
    def _ntEx(self):
        """
        number of total Ex (prior to deflating)
        """
        return int(self._vntEx.prod())

    @property
    def _vntEy(self):
        """
        vector number of total Ey (prior to deflating)
        """
        return np.r_[self._ntNx, self.nCy, self._ntNz]

    @property
    def _ntEy(self):
        """
        number of total Ey (prior to deflating)
        """
        return int(self._vntEy.prod())

    @property
    def vnEy(self):
        """
        Number of y-edges in each direction

        Returns
        -------
        numpy.ndarray
            vnEy or None if dim < 2, (dim, )
        """
        if self.isSymmetric:
            return np.r_[self.nNx, self.nCy, self.nNz]
        return np.r_[self.nNx - 1, self.nCy, self.nNz]

    @property
    def _vntEz(self):
        """
        vector number of total Ez (prior to deflating)
        """
        return np.r_[self._ntNx, self._ntNy, self.nCz]

    @property
    def _ntEz(self):
        """
        number of total Ez (prior to deflating)
        """
        return int(self._vntEz.prod())

    @property
    def vnEz(self):
        """
        Returns
        -------
        numpy.ndarray
            Number of z-edges in each direction or None if nCy > 1, (dim, )
        """
        return np.r_[self.nNx, self.nNy, self.nCz]

    @property
    def nEz(self):
        """
        Returns
        -------
        int
            Number of z-edges
        """
        if self.isSymmetric:
            return self.vnEz.prod()
        return (np.r_[self.nNx-1, self.nNy, self.nCz]).prod() + self.nCz

    @property
    def vectorCCx(self):
        """Cell-centered grid vector (1D) in the x direction."""
        return np.r_[0, self.hx[:-1].cumsum()] + self.hx*0.5

    @property
    def vectorCCy(self):
        """Cell-centered grid vector (1D) in the y direction."""
        if self.isSymmetric:
            return np.r_[0, self.hy[:-1]]
        return np.r_[0, self.hy[:-1].cumsum()] + self.hy*0.5

    @property
    def vectorNx(self):
        """Nodal grid vector (1D) in the x direction."""
        if self.isSymmetric:
            return self.hx.cumsum()
        return np.r_[0, self.hx].cumsum()

    @property
    def _vectorNyFull(self):
        """
        full nodal y vector (prior to deflating)
        """
        if self.isSymmetric:
            return np.r_[0]
        return np.r_[0, self.hy.cumsum()]

    @property
    def vectorNy(self):
        """Nodal grid vector (1D) in the y direction."""
        # if self.isSymmetric:
        #     # There aren't really any nodes, but all the grids need
        #     # somewhere to live, why not zero?!
        #     return np.r_[0]
        return np.r_[0, self.hy[:-1].cumsum()]

    @property
    def _edgeExFull(self):
        """
        full x-edge lengths (prior to deflating)
        """
        return np.kron(
            np.ones(self._ntNz), np.kron(np.ones(self._ntNy), self.hx)
        )

    @property
    def edgeEx(self):
        """
        x-edge lengths - these are the radial edges. Radial edges only exist
        for a 3D cyl mesh.

        Returns
        -------
        numpy.ndarray
            vector of radial edge lengths
        """
        if getattr(self, '_edgeEx', None) is None:
            self._edgeEx = self._edgeExFull[~self._ishangingEx]
        return self._edgeEx

    @property
    def _edgeEyFull(self):
        """
        full vector of y-edge lengths (prior to deflating)
        """
        if self.isSymmetric:
            return 2*pi*self.gridN[:, 0]
        return np.kron(
            np.ones(self._ntNz),
            np.kron(self.hy, self.vectorNx)
        )

    @property
    def edgeEy(self):
        """
        y-edge lengths - these are the azimuthal edges. Azimuthal edges exist
        for all cylindrical meshes. These are arc-lengths (:math:`\\theta r`)

        Returns
        -------
        numpy.ndarray
            vector of the azimuthal edges
        """
        if getattr(self, '_edgeEy', None) is None:
            if self.isSymmetric:
                self._edgeEy = self._edgeEyFull
            else:
                self._edgeEy = self._edgeEyFull[~self._ishangingEy]
        return self._edgeEy

    @property
    def _edgeEzFull(self):
        """
        full z-edge lengths (prior to deflation)
        """
        return np.kron(
            self.hz,
            np.kron(np.ones(self._ntNy), np.ones(self._ntNx))
        )

    @property
    def edgeEz(self):
        """
        z-edge lengths - these are the vertical edges. Vertical edges only
        exist for a 3D cyl mesh.

        Returns
        -------
        numpy.ndarray
            vector of the vertical edges
        """
        if getattr(self, '_edgeEz', None) is None:

            self._edgeEz = self._edgeEzFull[~self._ishangingEz]
        return self._edgeEz

    @property
    def _edgeFull(self):
        """
        full edge lengths [r-edges, theta-edgesm z-edges] (prior to
        deflation)
        """
        if self.isSymmetric:
            raise NotImplementedError
        else:
            return np.r_[self._edgeExFull, self._edgeEyFull, self._edgeEzFull]

    @property
    def edge(self):
        """
        Edge lengths

        Returns
        -------
        numpy.ndarray
            vector of edge lengths :math:`(r, \\theta, z)`
        """
        if self.isSymmetric:
            return self.edgeEy
            # return 2*pi*self.gridN[:, 0]
        else:
            return np.r_[self.edgeEx, self.edgeEy, self.edgeEz]

    @property
    def _areaFxFull(self):
        """
        area of x-faces prior to deflation
        """
        if self.isSymmetric:
            return np.kron(self.hz, 2*pi*self.vectorNx)
        return np.kron(self.hz, np.kron(self.hy, self.vectorNx))

    @property
    def areaFx(self):
        """
        Area of the x-faces (radial faces). Radial faces exist on all
        cylindrical meshes

        .. math::
            A_x = r \\theta h_z

        Returns
        -------
        numpy.ndarray
            area of x-faces
        """
        if getattr(self, '_areaFx', None) is None:
            if self.isSymmetric:
                self._areaFx = self._areaFxFull
            else:
                self._areaFx = self._areaFxFull[~self._ishangingFx]
        return self._areaFx

    @property
    def _areaFyFull(self):
        """
        Area of y-faces (Azimuthal faces), prior to deflation.
        """
        return np.kron(self.hz, np.kron(np.ones(self._ntNy), self.hx))

    @property
    def areaFy(self):
        """
        Area of y-faces (Azimuthal faces). Azimuthal faces exist only on 3D
        cylindrical meshes.

        .. math::
            A_y = h_x h_z

        Returns
        -------
        numpy.ndarray
            area of y-faces
        """
        if getattr(self, '_areaFy', None) is None:
            if self.isSymmetric:
                raise Exception(
                    'There are no y-faces on the Cyl Symmetric mesh'
                )
            self._areaFy = self._areaFyFull[~self._ishangingFy]
        return self._areaFy

    @property
    def _areaFzFull(self):
        """
        area of z-faces prior to deflation
        """
        if self.isSymmetric:
            return np.kron(
                    np.ones_like(self.vectorNz), pi*(
                        self.vectorNx**2 -
                        np.r_[0, self.vectorNx[:-1]]**2
                    )
            )
        return np.kron(
            np.ones(self._ntNz), np.kron(
                self.hy,
                0.5 * (self.vectorNx[1:]**2 - self.vectorNx[:-1]**2)
            )
        )

    @property
    def areaFz(self):
        """
        Area of z-faces.

        .. math::
            A_z = \\frac{\\theta}{2} (r_2^2 - r_1^2)z

        Returns
        -------
        numpy.ndarray
            area of the z-faces
        """
        if getattr(self, '_areaFz', None) is None:
            if self.isSymmetric:
                self._areaFz = self._areaFzFull
            else:
                self._areaFz = self._areaFzFull[~self._ishangingFz]
        return self._areaFz

    @property
    def _areaFull(self):
        """
        Area of all faces (prior to delflation)
        """
        return np.r_[self._areaFxFull, self._areaFyFull, self._areaFzFull]

    @property
    def area(self):
        """
        Face areas

        For a 3D cyl mesh: [radial, azimuthal, vertical], while a cylindrically
        symmetric mesh doesn't have y-Faces, so it returns [radial, vertical]

        Returns
        -------
        numpy.ndarray
            face areas
        """
        # if getattr(self, '_area', None) is None:
        if self.isSymmetric:
            return np.r_[self.areaFx, self.areaFz]
        else:
            return np.r_[self.areaFx, self.areaFy, self.areaFz]

    @property
    def vol(self):
        """
        Volume of each cell

        Returns
        -------
        numpy.ndarray
            cell volumes
        """
        if getattr(self, '_vol', None) is None:
            if self.isSymmetric:
                az = pi*(self.vectorNx**2 - np.r_[0, self.vectorNx[:-1]]**2)
                self._vol = np.kron(self.hz, az)
            else:
                self._vol = np.kron(
                    self.hz, np.kron(
                        self.hy,
                        0.5 * (self.vectorNx[1:]**2 - self.vectorNx[:-1]**2)
                    )
                )
        return self._vol

    ###########################################################################
    # Active and Hanging Edges and Faces
    #
    #    To find the active edges, faces, we use krons of bools (sorry). It is
    #    more efficient than working with 3D matrices. For example...
    #
    #    The computation of `ishangingFx` (is the Fx face hanging? a vector of
    #    True and False corresponding to each face) can be computed using krons
    #    of bools:
    #
    #          hang_x = np.zeros(self._ntNx, dtype=bool)
    #          hang_x[0] = True
    #          ishangingFxBool = np.kron(
    #              np.ones(self.nCz, dtype=bool),  # 1 * 0 == 0
    #              np.kron(np.ones(self.nCy, dtype=bool), hang_x)
    #          )
    #          return self._ishangingFxBool
    #
    #
    #   This is equivalent to forming the 3D matrix and indexing the
    #   corresponding rows and columns (here, the hanging faces are all of
    #   the first x-faces along the axis of symmetry):
    #
    #         hang_x = np.zeros(self._vntFx, dtype=bool)
    #         hang_x[0, :, :] = True
    #         isHangingFxBool = mkvc(hang_x)
    #
    #
    # but krons of bools is more efficient.
    #
    ###########################################################################

    @property
    def _ishangingFx(self):
        """
        bool vector indicating if an x-face is hanging or not
        """
        if getattr(self, '_ishangingFxBool', None) is None:

            # the following is equivalent to
            #     hang_x = np.zeros(self._vntFx, dtype=bool)
            #     hang_x[0, :, :] = True
            #     isHangingFxBool = mkvc(hang_x)
            #
            # but krons of bools is more efficient

            hang_x = np.zeros(self._ntNx, dtype=bool)
            hang_x[0] = True
            self._ishangingFxBool = np.kron(
                np.ones(self.nCz, dtype=bool),  # 1 * 0 == 0
                np.kron(
                    np.ones(self.nCy, dtype=bool),
                    hang_x
                )
            )
        return self._ishangingFxBool

    @property
    def _hangingFx(self):
        """
        dictionary of the indices of the hanging x-faces (keys) and a list
        of indices that the eliminated faces map to (if applicable)
        """
        if getattr(self, '_hangingFxDict', None) is None:
            self._hangingFxDict = dict(zip(
                np.nonzero(self._ishangingFx)[0].tolist(), [None]*self._nhFx
            ))
        return self._hangingFxDict

    @property
    def _ishangingFy(self):
        """
        bool vector indicating if a y-face is hanging or not
        """

        if getattr(self, '_ishangingFyBool', None) is None:
            hang_y = np.zeros(self._ntNy, dtype=bool)
            hang_y[-1] = True
            self._ishangingFyBool = np.kron(
                np.ones(self.nCz, dtype=bool),
                np.kron(
                    hang_y,
                    np.ones(self.nCx, dtype=bool)
                )
            )
        return self._ishangingFyBool

    @property
    def _hangingFy(self):
        """
        dictionary of the indices of the hanging y-faces (keys) and a list
        of indices that the eliminated faces map to (if applicable)
        """
        if getattr(self, '_hangingFyDict', None) is None:
            deflate_y = np.zeros(self._ntNy, dtype=bool)
            deflate_y[0] = True
            deflateFy = np.nonzero(np.kron(
                np.ones(self.nCz, dtype=bool),
                np.kron(
                    deflate_y,
                    np.ones(self.nCx, dtype=bool)
                )
            ))[0].tolist()
            self._hangingFyDict = dict(zip(
                np.nonzero(self._ishangingFy)[0].tolist(),
                deflateFy)
            )
        return self._hangingFyDict

    @property
    def _ishangingFz(self):
        """
        bool vector indicating if a z-face is hanging or not
        """
        if getattr(self, '_ishangingFzBool', None) is None:
            self._ishangingFzBool = np.zeros(self._ntFz, dtype=bool)
        return self._ishangingFzBool

    @property
    def _hangingFz(self):
        """
        dictionary of the indices of the hanging z-faces (keys) and a list
        of indices that the eliminated faces map to (if applicable)
        """
        return {}

    @property
    def _ishangingEx(self):
        """
        bool vector indicating if a x-edge is hanging or not
        """
        if getattr(self, '_ishangingExBool', None) is None:
            hang_y = np.zeros(self._ntNy, dtype=bool)
            hang_y[-1] = True
            self._ishangingExBool = np.kron(
                np.ones(self._ntNz, dtype=bool),
                np.kron(
                    hang_y,
                    np.ones(self.nCx, dtype=bool)
                )
            )
        return self._ishangingExBool

    @property
    def _hangingEx(self):
        """
        dictionary of the indices of the hanging x-edges (keys) and a list
        of indices that the eliminated faces map to (if applicable)
        """
        if getattr(self, '_hangingExDict', None) is None:
            deflate_y = np.zeros(self._ntNy, dtype=bool)
            deflate_y[0] = True
            deflateEx = np.nonzero(np.kron(
                np.ones(self._ntNz, dtype=bool),
                np.kron(
                    deflate_y,
                    np.ones(self.nCx, dtype=bool)
                )
            ))[0].tolist()
            self._hangingExDict = dict(zip(
                np.nonzero(self._ishangingEx)[0].tolist(), deflateEx
            ))
        return self._hangingExDict

    @property
    def _ishangingEy(self):
        """
        bool vector indicating if a y-edge is hanging or not
        """
        if getattr(self, '_ishangingEyBool', None) is None:
            hang_x = np.zeros(self._ntNx, dtype=bool)
            hang_x[0] = True
            self._ishangingEyBool = np.kron(
                np.ones(self._ntNz, dtype=bool),
                np.kron(
                    np.ones(self.nCy, dtype=bool),
                    hang_x
                )
            )
        return self._ishangingEyBool

    @property
    def _hangingEy(self):
        """
        dictionary of the indices of the hanging y-edges (keys) and a list
        of indices that the eliminated faces map to (if applicable)
        """
        if getattr(self, '_hangingEyDict', None) is None:
            self._hangingEyDict = dict(zip(
                np.nonzero(self._ishangingEy)[0].tolist(),
                [None]*len(self._ishangingEyBool))
            )
        return self._hangingEyDict

    @property
    def _axis_of_symmetry_Ez(self):
        """
        bool vector indicating if a z-edge is along the axis of symmetry or not
        """
        if getattr(self, '_axis_of_symmetry_EzBool', None) is None:
            axis_x = np.zeros(self._ntNx, dtype=bool)
            axis_x[0] = True

            axis_y = np.zeros(self._ntNy, dtype=bool)
            axis_y[0] = True
            self._axis_of_symmetry_EzBool = np.kron(
                np.ones(self.nCz, dtype=bool),
                np.kron(
                    axis_y,
                    axis_x
                )
            )
        return self._axis_of_symmetry_EzBool

    @property
    def _ishangingEz(self):
        """
        bool vector indicating if a z-edge is hanging or not
        """
        if getattr(self, '_ishangingEzBool', None) is None:
            if self.isSymmetric:
                self._ishangingEzBool = np.ones(self._ntEz, dtype=bool)
            else:
                hang_x = np.zeros(self._ntNx, dtype=bool)
                hang_x[0] = True

                hang_y = np.zeros(self._ntNy, dtype=bool)
                hang_y[-1] = True

                hangingEz = np.kron(
                    np.ones(self.nCz, dtype=bool),
                    (
                        # True * False = False
                        np.kron(
                            np.ones(self._ntNy, dtype=bool),
                            hang_x
                        ) |
                        np.kron(
                            hang_y,
                            np.ones(self._ntNx, dtype=bool)
                        )
                    )
                )

                self._ishangingEzBool = hangingEz & ~self._axis_of_symmetry_Ez

        return self._ishangingEzBool

    @property
    def _hangingEz(self):
        """
        dictionary of the indices of the hanging z-edges (keys) and a list
        of indices that the eliminated faces map to (if applicable)
        """
        if getattr(self, '_hangingEzDict', None) is None:
            # deflate
            deflateEz = np.hstack([
                np.hstack([
                    np.zeros(self._ntNy-1, dtype=int),
                    np.arange(1, self._ntNx, dtype=int)
                ]) +
                i*int(self._ntNx*self._ntNy)
                for i in range(self.nCz)
            ])
            deflate = zip(
                np.nonzero(self._ishangingEz)[0].tolist(), deflateEz
            )

            self._hangingEzDict = dict(deflate)
        return self._hangingEzDict

    @property
    def _axis_of_symmetry_N(self):
        """
        bool vector indicating if a node is along the axis of symmetry or not
        """
        if getattr(self, '_axis_of_symmetry_NBool', None) is None:
            axis_x = np.zeros(self._ntNx, dtype=bool)
            axis_x[0] = True

            axis_y = np.zeros(self._ntNy, dtype=bool)
            axis_y[0] = True
            self._axis_of_symmetry_NBool = np.kron(
                np.ones(self._ntNz, dtype=bool),
                np.kron(
                    axis_y,
                    axis_x
                )
            )
        return self._axis_of_symmetry_NBool

    @property
    def _ishangingN(self):
        """
        bool vector indicating if a node is hanging or not
        """
        if getattr(self, '_ishangingNBool', None) is None:
            hang_x = np.zeros(self._ntNx, dtype=bool)
            hang_x[0] = True

            hang_y = np.zeros(self._ntNy, dtype=bool)
            hang_y[-1] = True

            hangingN = np.kron(
                np.ones(self._ntNz, dtype=bool),
                (
                    np.kron(
                        np.ones(self._ntNy, dtype=bool),
                        hang_x
                    ) |
                    np.kron(
                        hang_y,
                        np.ones(self._ntNx, dtype=bool)
                    )
                )
            )

            self._ishangingNBool = hangingN & ~self._axis_of_symmetry_N

        return self._ishangingNBool

    @property
    def _hangingN(self):
        """
        dictionary of the indices of the hanging nodes (keys) and a list
        of indices that the eliminated faces map to (if applicable)
        """
        if getattr(self, '_hangingNDict', None) is None:
            # go by layer
            deflateN = np.hstack([
                np.hstack([
                    np.zeros(self._ntNy-1, dtype=int),
                    np.arange(1, self._ntNx, dtype=int)
                ]) +
                i*int(self._ntNx*self._ntNy)
                for i in range(self._ntNz)
            ]).tolist()
            self._hangingNDict = dict(zip(
                np.nonzero(self._ishangingN)[0].tolist(), deflateN
            ))
        return self._hangingNDict

    ####################################################
    # Grids
    ####################################################

    @property
    def _gridNFull(self):
        """
        Full Nodal grid (including hanging nodes)
        """
        return ndgrid([
            self.vectorNx, self._vectorNyFull, self.vectorNz
        ])

    @property
    def gridN(self):
        """
        Nodal grid in cylindrical coordinates :math:`(r, \\theta, z)`.
        Nodes do not exist in a cylindrically symmetric mesh.

        Returns
        -------
        numpy.ndarray
            grid locations of nodes
        """
        if self.isSymmetric:
            self._gridN = self._gridNFull
        if getattr(self, '_gridN', None) is None:
            self._gridN = self._gridNFull[~self._ishangingN, :]
        return self._gridN

    @property
    def _gridFxFull(self):
        """
        Full Fx grid (including hanging faces)
        """
        return ndgrid([
            self.vectorNx, self.vectorCCy, self.vectorCCz
        ])

    @property
    def gridFx(self):
        """
        Grid of x-faces (radial-faces) in cylindrical coordinates
        :math:`(r, \\theta, z)`.

        Returns
        -------
        numpy.ndarray
            grid locations of radial faces
        """
        if getattr(self, '_gridFx', None) is None:
            if self.isSymmetric:
                return super(CylMesh, self).gridFx
            else:
                self._gridFx = self._gridFxFull[~self._ishangingFx, :]
        return self._gridFx

    @property
    def _gridEyFull(self):
        """
        Full grid of y-edges (including eliminated edges)
        """
        return super(CylMesh, self).gridEy

    @property
    def gridEy(self):
        """
        Grid of y-edges (azimuthal-faces) in cylindrical coordinates
        :math:`(r, \\theta, z)`.

        Returns
        -------
        numpy.ndarray
            grid locations of azimuthal faces
        """
        if getattr(self, '_gridEy', None) is None:
            if self.isSymmetric:
                return self._gridEyFull
            else:
                self._gridEy = self._gridEyFull[~self._ishangingEy, :]
        return self._gridEy

    @property
    def _gridEzFull(self):
        """
        Full z-edge grid (including hanging edges)
        """
        return ndgrid([
            self.vectorNx, self._vectorNyFull, self.vectorCCz
        ])

    @property
    def gridEz(self):
        """
        Grid of z-faces (vertical-faces) in cylindrical coordinates
        :math:`(r, \\theta, z)`.

        Returns
        -------
        numpy.ndarray
            grid locations of radial faces
        """
        if getattr(self, '_gridEz', None) is None:
            if self.isSymmetric:
                self._gridEz = None
            else:
                self._gridEz = self._gridEzFull[~self._ishangingEz, :]
        return self._gridEz

    ####################################################
    # Operators
    ####################################################

    @property
    def faceDiv(self):
        """
        Construct divergence operator (faces to cell-centres).
        """
        if getattr(self, '_faceDiv', None) is None:
            # Compute faceDivergence operator on faces
            D1 = self.faceDivx
            D3 = self.faceDivz
            if self.isSymmetric:
                D = sp.hstack((D1, D3), format="csr")
            elif self.nCy > 1:
                D2 = self.faceDivy
                D = sp.hstack((D1, D2, D3), format="csr")
            self._faceDiv = D
        return self._faceDiv

    @property
    def faceDivx(self):
        """
        Construct divergence operator in the x component
        (faces to cell-centres).
        """
        if getattr(self, '_faceDivx', None) is None:
            if self.isSymmetric:
                D1 = kron3(
                    speye(self.nCz), speye(self.nCy),
                    ddx(self.nCx)[:, 1:]
                )
            else:
                D1 = super(CylMesh, self)._faceDivStencilx

            S = self._areaFxFull
            V = self.vol
            self._faceDivx = sdiag(1/V)*D1*sdiag(S)

            if not self.isSymmetric:
                self._faceDivx = (
                    self._faceDivx *
                    self._deflationMatrix(
                        'Fx', asOnes=True
                    ).T
                )

        return self._faceDivx

    @property
    def faceDivy(self):
        """
        Construct divergence operator in the y component
        (faces to cell-centres).
        """
        if getattr(self, '_faceDivy', None) is None:
            D2 = super(CylMesh, self)._faceDivStencily
            S = self._areaFyFull  # self.r(self.area, 'F', 'Fy', 'V')
            V = self.vol
            self._faceDivy = (
                sdiag(1/V)*D2*sdiag(S) *
                self._deflationMatrix('Fy', asOnes=True).T
            )
        return self._faceDivy

    @property
    def faceDivz(self):
        """
        Construct divergence operator in the z component
        (faces to cell-centres).
        """
        if getattr(self, '_faceDivz', None) is None:
            D3 = super(CylMesh, self)._faceDivStencilz
            S = self._areaFzFull
            V = self.vol
            self._faceDivz = sdiag(1/V)*D3*sdiag(S)
        return self._faceDivz

    # @property
    # def _cellGradxStencil(self):
    #     n = self.vnC

    #     if self.isSymmetric:
    #         G1 = sp.kron(speye(n[2]), ddxCellGrad(n[0], BC))
    #     else:
    #         G1 = self._deflationMatrix('Fx').T * kron3(
    #             speye(n[2]), speye(n[1]), ddxCellGrad(n[0], BC)
    #         )
    #     return G1

    @property
    def cellGradx(self):
        raise NotImplementedError
        # if getattr(self, '_cellGradx', None) is None:
        #     G1 = super(CylMesh, self)._cellGradxStencil
        #     V = self._deflationMatrix('F', withHanging='True', asOnes='True')*self.aveCC2F*self.vol
        #     A = self.area
        #     L = (A/V)[:self._ntFx]
        #     # L = self.r(L, 'F', 'Fx', 'V')
        #     # L = A[:self.nFx] / V
        #     self._cellGradx = self._deflationMatrix('Fx')*sdiag(L)*G1
        # return self._cellGradx

    @property
    def _cellGradyStencil(self):
        raise NotImplementedError

    @property
    def _cellGradzStencil(self):
        raise NotImplementedError

    @property
    def _cellGradStencil(self):
        raise NotImplementedError

    @property
    def cellGrad(self):
        """The cell centered Gradient, takes you to cell faces."""
        raise NotImplementedError('Cell Grad is not yet implemented.')

    # @property
    # def _nodalGradStencilx(self):
    #     if self.isSymmetric:
    #         return None
    #     return kron3(speye(self.nNz), speye(self.nNy), ddx(self.nCx))

    # @property
    # def _nodalGradStencily(self):
    #     if self.isSymmetric:
    #         None
    #         # return kron3(speye(self.nNz), ddx(self.nCy), speye(self.nNx)) * self._deflationMatrix('Ey')
    #     return kron3(speye(self.nNz), ddx(self.nCy), speye(self.nNx))

    # @property
    # def _nodalGradStencilz(self):
    #     if self.isSymmetric:
    #         return None
    #     return kron3(ddx(self.nCz), speye(self.nNy), speye(self.nNx))

    # @property
    # def _nodalGradStencil(self):
    #     if self.isSymmetric:
    #         return None
    #     else:
    #         G = self._deflationMatrix('E').T * sp.vstack((
    #             self._nodalGradStencilx,
    #             self._nodalGradStencily,
    #             self._nodalGradStencilz
    #         ), format="csr") * self._deflationMatrix('N')
    #     return G

    @property
    def nodalGrad(self):
        """Construct gradient operator (nodes to edges)."""
        if self.isSymmetric:
            return None
        raise NotImplementedError('nodalGrad not yet implemented')

    @property
    def nodalLaplacian(self):
        """Construct laplacian operator (nodes to edges)."""
        raise NotImplementedError('nodalLaplacian not yet implemented')

    @property
    def edgeCurl(self):
        """
        The edgeCurl (edges to faces)

        Returns
        -------
        scipy.sparse.csr_matrix
            edge curl operator
        """
        if getattr(self, '_edgeCurl', None) is None:
            A = self.area
            E = self.edge

            if self.isSymmetric:
                # 1D Difference matricies
                dr = sp.spdiags(
                    (np.ones((self.nCx+1, 1))*[-1, 1]).T, [-1, 0],
                    self.nCx, self.nCx, format="csr"
                )
                dz = sp.spdiags(
                    (np.ones((self.nCz+1, 1))*[-1, 1]).T, [0, 1],
                    self.nCz, self.nCz+1, format="csr"
                )
                # 2D Difference matricies
                Dr = sp.kron(sp.identity(self.nNz), dr)
                Dz = -sp.kron(dz, sp.identity(self.nCx))

                # Edge curl operator
                self._edgeCurl = (
                    sdiag(1/A)*sp.vstack((Dz, Dr)) * sdiag(E)
                )
            else:
                self._edgeCurl = (
                    sdiag(1/self.area) *
                    self._deflationMatrix('F', asOnes=False) *
                    self._edgeCurlStencil *
                    sdiag(self._edgeFull) *
                    self._deflationMatrix('E', asOnes=True).T
                )

        return self._edgeCurl

    @property
    def aveEx2CC(self):
        """
        averaging operator of x-edges (radial) to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from x-edges to cell centers
        """
        if self.isSymmetric:
            raise Exception('There are no x-edges on a cyl symmetric mesh')
        return kron3(
            av(self.vnC[2]),
            av(self.vnC[1]),
            speye(self.vnC[0])
        ) * self._deflationMatrix('Ex', asOnes=True).T

    @property
    def aveEy2CC(self):
        """
        averaging operator of y-edges (azimuthal) to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from y-edges to cell centers
        """
        if self.isSymmetric:
            avR = av(self.vnC[0])[:, 1:]
            return sp.kron(av(self.vnC[2]), avR, format="csr")
        else:
            return kron3(
                av(self.vnC[2]),
                speye(self.vnC[1]),
                av(self.vnC[0])
            )*self._deflationMatrix('Ey', asOnes=True).T

    @property
    def aveEz2CC(self):
        """
        averaging operator of z-edges to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from z-edges to cell centers
        """
        if self.isSymmetric:
            raise Exception('There are no z-edges on a cyl symmetric mesh')
        return kron3(
            speye(self.vnC[2]),
            av(self.vnC[1]),
            av(self.vnC[0])
        ) * self._deflationMatrix('Ez', asOnes=True).T

    @property
    def aveE2CC(self):
        """
        averaging operator of edges to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from edges to cell centers
        """
        if getattr(self, '_aveE2CC', None) is None:
            # The number of cell centers in each direction
            # n = self.vnC
            if self.isSymmetric:
                self._aveE2CC = self.aveEy2CC
            else:
                self._aveE2CC = 1./self.dim * sp.hstack(
                    (self.aveEx2CC, self.aveEy2CC, self.aveEz2CC),
                    format="csr"
                )
        return self._aveE2CC

    @property
    def aveE2CCV(self):
        """
        averaging operator of edges to a cell centered vector

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from edges to cell centered vectors
        """
        if self.isSymmetric:
            return self.aveE2CC
        else:
            if getattr(self, '_aveE2CCV', None) is None:
                self._aveE2CCV = sp.block_diag(
                    (self.aveEx2CC, self.aveEy2CC, self.aveEz2CC),
                    format="csr"
                )
        return self._aveE2CCV

    @property
    def aveFx2CC(self):
        """
        averaging operator of x-faces (radial) to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from x-faces to cell centers
        """
        avR = av(self.vnC[0])[:, 1:]  # TODO: this should be handled by a deflation matrix
        return kron3(
            speye(self.vnC[2]), speye(self.vnC[1]), avR
        )

    @property
    def aveFy2CC(self):
        """
        averaging operator of y-faces (azimuthal) to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from y-faces to cell centers
        """
        return kron3(
            speye(self.vnC[2]), av(self.vnC[1]),
            speye(self.vnC[0])
        ) * self._deflationMatrix('Fy', asOnes=True).T

    @property
    def aveFz2CC(self):
        """
        averaging operator of z-faces (vertical) to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from z-faces to cell centers
        """

        return kron3(
            av(self.vnC[2]), speye(self.vnC[1]),
            speye(self.vnC[0])
        )

    @property
    def aveF2CC(self):
        """
        averaging operator of faces to cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from faces to cell centers
        """
        if getattr(self, '_aveF2CC', None) is None:
            n = self.vnC
            if self.isSymmetric:
                self._aveF2CC = 0.5*(
                    sp.hstack((self.aveFx2CC, self.aveFz2CC), format="csr")
                )
            else:
                self._aveF2CC = 1./self.dim*(
                    sp.hstack(
                        (self.aveFx2CC, self.aveFy2CC, self.aveFz2CC),
                        format="csr"
                    )
                )
        return self._aveF2CC

    @property
    def aveF2CCV(self):
        """
        averaging operator of x-faces (radial) to cell centered vectors

        Returns
        -------
        scipy.sparse.csr_matrix
            matrix that averages from faces to cell centered vectors
        """
        if getattr(self, '_aveF2CCV', None) is None:
            # n = self.vnC
            if self.isSymmetric:
                self._aveF2CCV = sp.block_diag(
                    (self.aveFx2CC, self.aveFz2CC), format="csr"
                )
            else:
                self._aveF2CCV = sp.block_diag(
                    (self.aveFx2CC, self.aveFy2CC, self.aveFz2CC),
                    format="csr"
                )
        return self._aveF2CCV

    ####################################################
    # Deflation Matrices
    ####################################################

    def _deflationMatrix(self, location, asOnes=False):
        """
        construct the deflation matrix to remove hanging edges / faces / nodes
        from the operators
        """
        if location not in [
            'N', 'F', 'Fx', 'Fy', 'Fz', 'E', 'Ex', 'Ey', 'Ez', 'CC'
        ]:
            raise AssertionError(
                'Location must be a grid location, not {}'.format(location)
            )
        if location == 'CC':
            return speye(self.nC)

        elif location in ['E', 'F']:
            if self.isSymmetric:
                if location == 'E':
                    return self._deflationMatrix('Ey', asOnes=asOnes)
                elif location == 'F':
                    return sp.block_diag([
                        self._deflationMatrix(location+coord, asOnes=asOnes)
                        for coord in ['x', 'z']
                    ])
            return sp.block_diag([
                self._deflationMatrix(location+coord, asOnes=asOnes)
                for coord in ['x', 'y', 'z']
            ])

        R = speye(getattr(self, '_nt{}'.format(location)))
        hanging_dict = getattr(self, '_hanging{}'.format(location))
        nothanging = ~getattr(self, '_ishanging{}'.format(location))

        # remove eliminated edges / faces (eg. Fx just doesn't exist)
        hang = {k: v for k, v in hanging_dict.items() if v is not None}

        values = list(hang.values())
        entries = np.ones(len(values))

        if not asOnes and len(hang) > 0:
            repeats = set(values)
            repeat_locs = [
                (np.r_[values] == repeat).nonzero()[0]
                for repeat in repeats
            ]
            for loc in repeat_locs:
                entries[loc] = 1./len(loc)

        Hang = sp.csr_matrix(
            (entries, (values, list(hang.keys()))),
            shape=(
                getattr(self, '_nt{}'.format(location)),
                getattr(self, '_nt{}'.format(location))
            )
        )
        R = R + Hang

        R = R[nothanging, :]

        if not asOnes:
            R = sdiag(1./R.sum(1)) * R

        return R

    ####################################################
    # Interpolation
    ####################################################

    def getInterpolationMat(self, loc, locType='CC', zerosOutside=False):
        """ Produces interpolation matrix

        Parameters
        ----------
        loc : numpy.ndarray
            Location of points to interpolate to

        locType : str
            What to interpolate locType can be::

            'Ex'    -> x-component of field defined on edges
            'Ey'    -> y-component of field defined on edges
            'Ez'    -> z-component of field defined on edges
            'Fx'    -> x-component of field defined on faces
            'Fy'    -> y-component of field defined on faces
            'Fz'    -> z-component of field defined on faces
            'N'     -> scalar field defined on nodes
            'CC'    -> scalar field defined on cell centers
            'CCVx'  -> x-component of vector field defined on cell centers
            'CCVy'  -> y-component of vector field defined on cell centers
            'CCVz'  -> z-component of vector field defined on cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            M, the interpolation matrix

        """
        if self.isSymmetric and locType in ['Ex', 'Ez', 'Fy']:
            raise Exception(
                "Symmetric CylMesh does not support {0!s} interpolation, "
                "as this variable does not exist.".format(locType)
            )

        if locType in ['CCVx', 'CCVy', 'CCVz']:
            Q = interpmat(loc, *self.getTensor('CC'))
            Z = spzeros(loc.shape[0], self.nC)
            if locType == 'CCVx':
                Q = sp.hstack([Q, Z])
            elif locType == 'CCVy':
                Q = sp.hstack([Q])
            elif locType == 'CCVz':
                Q = sp.hstack([Z, Q])

            if zerosOutside:
                Q[indZeros, :] = 0

            return Q.tocsr()

        return self._getInterpolationMat(loc, locType, zerosOutside)

    def cartesianGrid(self, locType='CC', theta_shift=None):
        """
        Takes a grid location ('CC', 'N', 'Ex', 'Ey', 'Ez', 'Fx', 'Fy', 'Fz')
        and returns that grid in cartesian coordinates

        Parameters
        ----------
        locType : str
            grid location

        Returns
        -------
        numpy.ndarray
            cartesian coordinates for the cylindrical grid
        """
        grid = getattr(self, 'grid{}'.format(locType)).copy()
        if theta_shift is not None:
            grid[:, 1] = grid[:, 1] - theta_shift
        return cyl2cart(grid)  # TODO: account for cartesian origin

    def getInterpolationMatCartMesh(self, Mrect, locType='CC', locTypeTo=None):
        """
        Takes a cartesian mesh and returns a projection to translate onto
        the cartesian grid.

        Parameters
        ----------
        Mrect : discretize.base.BaseMesh
            the mesh to interpolate on to

        locType : str
            grid location ('CC', 'N', 'Ex', 'Ey', 'Ez', 'Fx', 'Fy', 'Fz')

        locTypeTo : str
            grid location to interpolate to. If None, the same grid type as `locType` will be assumed

        Returns
        -------
        scipy.sparse.csr_matrix
            M, the interpolation matrix
        """

        if not self.isSymmetric:
            raise AssertionError(
                "Currently we have not taken into account other projections "
                "for more complicated CylMeshes"
            )

        if locTypeTo is None:
            locTypeTo = locType

        if locType == 'F':
            # do this three times for each component
            X = self.getInterpolationMatCartMesh(
                Mrect, locType='Fx', locTypeTo=locTypeTo+'x'
            )
            Y = self.getInterpolationMatCartMesh(
                Mrect, locType='Fy', locTypeTo=locTypeTo+'y'
            )
            Z = self.getInterpolationMatCartMesh(
                Mrect, locType='Fz', locTypeTo=locTypeTo+'z'
            )
            return sp.vstack((X, Y, Z))
        if locType == 'E':
            X = self.getInterpolationMatCartMesh(
                Mrect, locType='Ex', locTypeTo=locTypeTo+'x'
            )
            Y = self.getInterpolationMatCartMesh(
                Mrect, locType='Ey', locTypeTo=locTypeTo+'y'
            )
            Z = spzeros(getattr(Mrect, 'n' + locTypeTo + 'z'), self.nE)
            return sp.vstack((X, Y, Z))

        grid = getattr(Mrect, 'grid' + locTypeTo)
        # This is unit circle stuff, 0 to 2*pi, starting at x-axis, rotating
        # counter clockwise in an x-y slice
        theta = - np.arctan2(
            grid[:, 0] - self.cartesianOrigin[0], grid[:, 1] -
            self.cartesianOrigin[1]
        ) + np.pi/2
        theta[theta < 0] += np.pi*2.0
        r = ((grid[:, 0] - self.cartesianOrigin[0])**2 + (grid[:, 1] -
             self.cartesianOrigin[1])**2)**0.5

        if locType in ['CC', 'N', 'Fz', 'Ez']:
            G, proj = np.c_[r, theta, grid[:, 2]], np.ones(r.size)
        else:
            dotMe = {
                'Fx': Mrect.normals[:Mrect.nFx, :],
                'Fy': Mrect.normals[Mrect.nFx:(Mrect.nFx + Mrect.nFy), :],
                'Fz': Mrect.normals[-Mrect.nFz:, :],
                'Ex': Mrect.tangents[:Mrect.nEx, :],
                'Ey': Mrect.tangents[Mrect.nEx:(Mrect.nEx+Mrect.nEy), :],
                'Ez': Mrect.tangents[-Mrect.nEz:, :],
            }[locTypeTo]
            if 'F' in locType:
                normals = np.c_[
                    np.cos(theta), np.sin(theta), np.zeros(theta.size)
                ]
                proj = (normals * dotMe).sum(axis=1)
            if 'E' in locType:
                tangents = np.c_[
                    -np.sin(theta), np.cos(theta), np.zeros(theta.size)
                ]
                proj = (tangents * dotMe).sum(axis=1)
            G = np.c_[r, theta, grid[:, 2]]

        interpType = locType
        if interpType == 'Fy':
            interpType = 'Fx'
        elif interpType == 'Ex':
            interpType = 'Ey'

        Pc2r = self.getInterpolationMat(G, interpType)
        Proj = sdiag(proj)
        return Proj * Pc2r



from __future__ import print_function
import numpy as np
import properties
from properties.math import TYPE_MAPPINGS

from discretize import utils
from .base import BaseRectangularMesh
from discretize.DiffOperators import DiffOperators
from discretize.InnerProducts import InnerProducts
from discretize.View import CurviView


# Some helper functions.
def length2D(x):
    return (x[:, 0]**2 + x[:, 1]**2)**0.5


def length3D(x):
    return (x[:, 0]**2 + x[:, 1]**2 + x[:, 2]**2)**0.5


def normalize2D(x):
    return x/np.kron(np.ones((1, 2)), utils.mkvc(length2D(x), 2))


def normalize3D(x):
    return x/np.kron(np.ones((1, 3)), utils.mkvc(length3D(x), 2))


class CurvilinearMesh(
    BaseRectangularMesh, DiffOperators, InnerProducts, CurviView
):
    """CurvilinearMesh is a mesh class that deals with curvilinear meshes.

    Example of a curvilinear mesh:

    .. plot::
        :include-source:

        import discretize
        X, Y = discretize.utils.exampleLrmGrid([3,3],'rotate')
        mesh = discretize.CurvilinearMesh([X, Y])
        mesh.plotGrid(show_it=True)
    """

    _meshType = 'Curv'

    nodes = properties.List(
        "List of arrays describing the node locations",
        prop=properties.Array(
            "node locations in an n-dimensional array",
            shape={('*', '*'), ('*', '*', '*')}
        ),
        min_length=2,
        max_length=3
    )

    def __init__(self, nodes=None, **kwargs):

        self.nodes = nodes

        if '_n' in kwargs.keys():
            n = kwargs.pop('_n')
            assert (n == np.array(self.nodes[0].shape)-1).all(), (
                "Unexpected n-values. {} was provided, {} was expected".format(
                    n, np.array(self.nodes[0].shape)-1
                )
            )
        else:
            n = np.array(self.nodes[0].shape)-1

        BaseRectangularMesh.__init__(self, n, **kwargs)

        # Save nodes to private variable _gridN as vectors
        self._gridN = np.ones((self.nodes[0].size, self.dim))
        for i, node_i in enumerate(self.nodes):
            self._gridN[:, i] = utils.mkvc(node_i.astype(float))

    @properties.validator('nodes')
    def _check_nodes(self, change):
        assert len(change['value']) > 1, "len(node) must be greater than 1"

        for i, change['value'][i] in enumerate(change['value']):
            assert change['value'][i].shape == change['value'][0].shape, (
                "change['value'][{0:d}] is not the same shape as "
                "change['value'][0]".format(i)
            )

        assert len(change['value'][0].shape) == len(change['value']), (
            "Dimension mismatch"
        )

    @property
    def gridCC(self):
        """
        Cell-centered grid
        """
        if getattr(self, '_gridCC', None) is None:
            self._gridCC = np.concatenate(
                [self.aveN2CC*self.gridN[:, i] for i in range(self.dim)]
            ).reshape((-1, self.dim), order='F')
        return self._gridCC

    @property
    def gridN(self):
        """
        Nodal grid.
        """
        if getattr(self, '_gridN', None) is None:
            raise Exception("Someone deleted this. I blame you.")
        return self._gridN

    @property
    def gridFx(self):
        """
        Face staggered grid in the x direction.
        """

        if getattr(self, '_gridFx', None) is None:
            N = self.r(self.gridN, 'N', 'N', 'M')
            if self.dim == 2:
                XY = [utils.mkvc(0.5 * (n[:, :-1] + n[:, 1:])) for n in N]
                self._gridFx = np.c_[XY[0], XY[1]]
            elif self.dim == 3:
                XYZ = [utils.mkvc(0.25 * (n[:, :-1, :-1] + n[:, :-1, 1:] +
                       n[:, 1:, :-1] + n[:, 1:, 1:])) for n in N]
                self._gridFx = np.c_[XYZ[0], XYZ[1], XYZ[2]]
        return self._gridFx

    @property
    def gridFy(self):
        """
        Face staggered grid in the y direction.
        """

        if getattr(self, '_gridFy', None) is None:
            N = self.r(self.gridN, 'N', 'N', 'M')
            if self.dim == 2:
                XY = [utils.mkvc(0.5 * (n[:-1, :] + n[1:, :])) for n in N]
                self._gridFy = np.c_[XY[0], XY[1]]
            elif self.dim == 3:
                XYZ = [utils.mkvc(0.25 * (n[:-1, :, :-1] + n[:-1, :, 1:] +
                       n[1:, :, :-1] + n[1:, :, 1:])) for n in N]
                self._gridFy = np.c_[XYZ[0], XYZ[1], XYZ[2]]
        return self._gridFy

    @property
    def gridFz(self):
        """
        Face staggered grid in the y direction.
        """

        if getattr(self, '_gridFz', None) is None:
            N = self.r(self.gridN, 'N', 'N', 'M')
            XYZ = [utils.mkvc(0.25 * (n[:-1, :-1, :] + n[:-1, 1:, :] +
                   n[1:, :-1, :] + n[1:, 1:, :])) for n in N]
            self._gridFz = np.c_[XYZ[0], XYZ[1], XYZ[2]]
        return self._gridFz

    @property
    def gridEx(self):
        """
        Edge staggered grid in the x direction.
        """
        if getattr(self, '_gridEx', None) is None:
            N = self.r(self.gridN, 'N', 'N', 'M')
            if self.dim == 2:
                XY = [utils.mkvc(0.5 * (n[:-1, :] + n[1:, :])) for n in N]
                self._gridEx = np.c_[XY[0], XY[1]]
            elif self.dim == 3:
                XYZ = [utils.mkvc(0.5 * (n[:-1, :, :] + n[1:, :, :])) for n in N]
                self._gridEx = np.c_[XYZ[0], XYZ[1], XYZ[2]]
        return self._gridEx

    @property
    def gridEy(self):
        """
        Edge staggered grid in the y direction.
        """
        if getattr(self, '_gridEy', None) is None:
            N = self.r(self.gridN, 'N', 'N', 'M')
            if self.dim == 2:
                XY = [utils.mkvc(0.5 * (n[:, :-1] + n[:, 1:])) for n in N]
                self._gridEy = np.c_[XY[0], XY[1]]
            elif self.dim == 3:
                XYZ = [utils.mkvc(0.5 * (n[:, :-1, :] + n[:, 1:, :])) for n in N]
                self._gridEy = np.c_[XYZ[0], XYZ[1], XYZ[2]]
        return self._gridEy

    @property
    def gridEz(self):
        """
        Edge staggered grid in the z direction.
        """
        if getattr(self, '_gridEz', None) is None and self.dim == 3:
            N = self.r(self.gridN, 'N', 'N', 'M')
            XYZ = [utils.mkvc(0.5 * (n[:, :, :-1] + n[:, :, 1:])) for n in N]
            self._gridEz = np.c_[XYZ[0], XYZ[1], XYZ[2]]
        return self._gridEz

    # --------------- Geometries ---------------------
    #
    #
    # ------------------- 2D -------------------------
    #
    #         node(i,j)          node(i,j+1)
    #              A -------------- B
    #              |                |
    #              |    cell(i,j)   |
    #              |        I       |
    #              |                |
    #             D -------------- C
    #         node(i+1,j)        node(i+1,j+1)
    #
    # ------------------- 3D -------------------------
    #
    #
    #             node(i,j,k+1)       node(i,j+1,k+1)
    #                 E --------------- F
    #                /|               / |
    #               / |              /  |
    #              /  |             /   |
    #       node(i,j,k)         node(i,j+1,k)
    #            A -------------- B     |
    #            |    H ----------|---- G
    #            |   /cell(i,j)   |   /
    #            |  /     I       |  /
    #            | /              | /
    #            D -------------- C
    #       node(i+1,j,k)      node(i+1,j+1,k)

    @property
    def vol(self):
        """
        Construct cell volumes of the 3D model as 1d array
        """

        if getattr(self, '_vol', None) is None:
            if self.dim == 2:
                A, B, C, D = utils.indexCube('ABCD', self.vnC+1)
                normal, area = utils.faceInfo(np.c_[self.gridN, np.zeros(
                                              (self.nN, 1))], A, B, C, D)
                self._vol = area
            elif self.dim == 3:
                # Each polyhedron can be decomposed into 5 tetrahedrons
                # However, this presents a choice so we may as well divide in
                # two ways and average.
                A, B, C, D, E, F, G, H = utils.indexCube('ABCDEFGH', self.vnC +
                                                         1)

                vol1 = (utils.volTetra(self.gridN, A, B, D, E) +  # cutted edge top
                        utils.volTetra(self.gridN, B, E, F, G) +  # cutted edge top
                        utils.volTetra(self.gridN, B, D, E, G) +  # middle
                        utils.volTetra(self.gridN, B, C, D, G) +  # cutted edge bottom
                        utils.volTetra(self.gridN, D, E, G, H))   # cutted edge bottom

                vol2 = (utils.volTetra(self.gridN, A, F, B, C) +  # cutted edge top
                        utils.volTetra(self.gridN, A, E, F, H) +  # cutted edge top
                        utils.volTetra(self.gridN, A, H, F, C) +  # middle
                        utils.volTetra(self.gridN, C, H, D, A) +  # cutted edge bottom
                        utils.volTetra(self.gridN, C, G, H, F))   # cutted edge bottom

                self._vol = (vol1 + vol2)/2
        return self._vol

    @property
    def area(self):
        """
        Area of the faces
        """
        if (getattr(self, '_area', None) is None or
            getattr(self, '_normals', None) is None):
            # Compute areas of cell faces
            if(self.dim == 2):
                xy = self.gridN
                A, B = utils.indexCube('AB', self.vnC+1, np.array([self.nNx,
                                       self.nCy]))
                edge1 = xy[B, :] - xy[A, :]
                normal1 = np.c_[edge1[:, 1], -edge1[:, 0]]
                area1 = length2D(edge1)
                A, D = utils.indexCube('AD', self.vnC+1, np.array([self.nCx,
                                       self.nNy]))
                # Note that we are doing A-D to make sure the normal points the
                # right way.
                # Think about it. Look at the picture. Normal points towards C
                # iff you do this.
                edge2 = xy[A, :] - xy[D, :]
                normal2 = np.c_[edge2[:, 1], -edge2[:, 0]]
                area2 = length2D(edge2)
                self._area = np.r_[utils.mkvc(area1), utils.mkvc(area2)]
                self._normals = [normalize2D(normal1), normalize2D(normal2)]

            elif(self.dim == 3):

                A, E, F, B = utils.indexCube('AEFB', self.vnC+1, np.array(
                                             [self.nNx, self.nCy, self.nCz]))
                normal1, area1 = utils.faceInfo(self.gridN, A, E, F, B,
                                                average=False,
                                                normalizeNormals=False)

                A, D, H, E = utils.indexCube('ADHE', self.vnC+1, np.array(
                                             [self.nCx, self.nNy, self.nCz]))
                normal2, area2 = utils.faceInfo(self.gridN, A, D, H, E,
                                                average=False,
                                                normalizeNormals=False)

                A, B, C, D = utils.indexCube('ABCD', self.vnC+1, np.array(
                                             [self.nCx, self.nCy, self.nNz]))
                normal3, area3 = utils.faceInfo(self.gridN, A, B, C, D,
                                                average=False,
                                                normalizeNormals=False)

                self._area = np.r_[utils.mkvc(area1), utils.mkvc(area2),
                                   utils.mkvc(area3)]
                self._normals = [normal1, normal2, normal3]
        return self._area

    @property
    def normals(self):
        """
        Face normals: calling this will average
        the computed normals so that there is one
        per face. This is especially relevant in
        3D, as there are up to 4 different normals
        for each face that will be different.

        To reshape the normals into a matrix and get the y component::

            NyX, NyY, NyZ = M.r(M.normals, 'F', 'Fy', 'M')
        """

        if getattr(self, '_normals', None) is None:
            self.area  # calling .area will create the face normals
        if self.dim == 2:
            return normalize2D(np.r_[self._normals[0], self._normals[1]])
        elif self.dim == 3:
            normal1 = (
                self._normals[0][0] + self._normals[0][1] +
                self._normals[0][2] + self._normals[0][3]
            )/4
            normal2 = (
                self._normals[1][0] + self._normals[1][1] +
                self._normals[1][2] + self._normals[1][3]
            )/4
            normal3 = (
                self._normals[2][0] + self._normals[2][1] +
                self._normals[2][2] + self._normals[2][3]
            )/4
            return normalize3D(np.r_[normal1, normal2, normal3])

    @property
    def edge(self):
        """Edge lengths"""
        if getattr(self, '_edge', None) is None:
            if(self.dim == 2):
                xy = self.gridN
                A, D = utils.indexCube('AD', self.vnC+1, np.array([self.nCx,
                                                                  self.nNy]))
                edge1 = xy[D, :] - xy[A, :]
                A, B = utils.indexCube('AB', self.vnC+1, np.array([self.nNx,
                                                                   self.nCy]))
                edge2 = xy[B, :] - xy[A, :]
                self._edge = np.r_[utils.mkvc(length2D(edge1)),
                                   utils.mkvc(length2D(edge2))]
                self._tangents = np.r_[edge1, edge2]/np.c_[self._edge,
                                                           self._edge]
            elif(self.dim == 3):
                xyz = self.gridN
                A, D = utils.indexCube('AD', self.vnC+1, np.array([self.nCx,
                                                                   self.nNy,
                                                                   self.nNz]))
                edge1 = xyz[D, :] - xyz[A, :]
                A, B = utils.indexCube('AB', self.vnC+1, np.array([self.nNx,
                                                                   self.nCy,
                                                                   self.nNz]))
                edge2 = xyz[B, :] - xyz[A, :]
                A, E = utils.indexCube('AE', self.vnC+1, np.array([self.nNx,
                                                                   self.nNy,
                                                                   self.nCz]))
                edge3 = xyz[E, :] - xyz[A, :]
                self._edge = np.r_[utils.mkvc(length3D(edge1)),
                                   utils.mkvc(length3D(edge2)),
                                   utils.mkvc(length3D(edge3))]
                self._tangents = (np.r_[edge1, edge2, edge3] /
                                  np.c_[self._edge, self._edge, self._edge])
            return self._edge
        return self._edge

    @property
    def tangents(self):
        """Edge tangents"""
        if getattr(self, '_tangents', None) is None:
            self.edge  # calling .edge will create the tangents
        return self._tangents
